
from asyncio.log import logger
import functools
import logging
from typing import Callable

logger = logging.getLogger(__name__)

def push_batch(engine, names, data, batch_size):
    for off in range(0, len(data), batch_size):
        batch = data[off:(off + batch_size)]
        logger.debug('Pushing dst-batch [{}:{}] of size {}'.format(off, min(len(batch), off+batch_size), len(batch)))
        engine.insert_batch(names, batch)
        logger.debug('Pushed dst-batch [{}:{}] of size {}'.format(off, min(len(batch), off+batch_size), len(batch)))

def pull_batch(engine, batch_size):
    logger.debug('Pulling src-batch')
    names, batch = engine.fetch_batch(batch_size)
    logger.debug('Pulled src-batch of size {}'.format(len(batch)))
    return names, batch

def run_pull_push(src_engine, dst_engine, src_batch_size = 1000, dst_batch_size = 1000):
    counter = 0
    while True:
        names, data = pull_batch(src_engine, src_batch_size)
        if data is None or len(data) == 0:
            return
        push_batch(dst_engine, names, data, dst_batch_size)
        counter += 1
        logging.info('Processed {} batch of size {}.'.format(counter, len(data)))

def full_copy(src_engine, dst_engine, config):
    """
    Copies full content of source target into destination target.
    NOTE: it does not truncate, clean or delete existing records from target! 
    """
    logger.info('Starting replication.')
    src_engine.begin_full_fetch(config['src'])
    dst_engine.begin_insert(config['dst'])        
    run_pull_push(src_engine, dst_engine, config['src'].get('batch_size', 1000), config['dst'].get('batch_size', 1000))        
    logger.info('Replication finished.')

def incremental_update(src_engine, dst_engine, config):
    """
    Copies increment (new data) from source target into destination target.
    It determines increment based on values of rid field in both source and target.
    NOTE: it does not remove obsolete rids -- it is job of following Transform-tool like DBT.
    """
    logger.debug('Making request to get <src> latest rid...')
    src_rid = src_engine.get_latest_rid(config['src'])
    if src_rid is None:
        raise NotImplemented

    logger.debug('Making request to get <dst> latest rid...')
    dst_rid = dst_engine.get_latest_rid(config['dst'])

    logger.info('Latest rids: <src>={}, <dst>={}'.format(src_rid, dst_rid))
    logger.info('Starting replication.')
    while not dst_rid or dst_rid < src_rid:
        src_engine.begin_incremental_fetch(config['src'], dst_rid)
        dst_engine.begin_insert(config['dst'])        
        run_pull_push(src_engine, dst_engine, config['src'].get('batch_size', 1000), config['dst'].get('batch_size', 1000))

        logger.info('Finished sync. Updating <dst> rid...')
        dst_rid = dst_engine.get_latest_rid(config['dst'])
        logger.info('Latest rids: <src>={} (old), <dst>={} (updated)'.format(src_rid, dst_rid))
    logger.info('Replication finished.')


import functools
import importlib
from sqlite3 import Cursor

from .engine_base import BaseEngine
from .. import add_engine_factory

class DBAPIEngine(BaseEngine):
    id = 'dbapi'
    
    @staticmethod
    def _make_query_from_config(config):
        if 'query' in config:
            return '({}) t'.format(config['query'])
        if 'schema' in config:
            return '{}.{}'.format(config['schema'], config['table'])
        return '{}'.format(config['table'])

    @staticmethod
    def _insert_pyformat(cursor, table, names, values):
        col_names = ','.join(names)
        val_names = ','.join(['%({})s'.format(x) for x in names])
        query = 'insert into {} ({}) values ({})'.format(table, col_names, val_names)
        cursor.executemany(query, [dict(zip(names, x)) for x in values])

        
    @staticmethod
    def _insert_pyformat_optimized(cursor, table, names, values, batch_size=1000):
        batch_rows = max(1, int(batch_size / len(names)))
        col_names = ','.join(names)
        row_tmplt = '({})'.format(','.join(['%s']*len(names)))
        query_top = 'insert into {} ({}) values '.format(table, col_names)
        for i in range(0, len(values), batch_rows):
            batch = values[i:min(i+batch_rows, len(values))]
            query = query_top + ','.join([row_tmplt]*len(batch))
            cursor.execute(query, [v for x in batch for v in x])
        
    def __init__(self, connection_config):     
        self.driver_name = connection_config['driver']
        self.driver = importlib.import_module(self.driver_name)
        if self.driver.paramstyle != 'pyformat':
            raise NotImplementedError('Support for drivers with paramstyle other than pyformat is not implemented!')
        
        driver_keywords = ['dsn', 'database', 'user', 'host', 'port', 'password']
        driver_params = {k: v for k,v in connection_config.items() if k in driver_keywords}
        driver_params.update(connection_config.get('driver-params', {}))
        self.conn = self.driver.connect(**driver_params)

        self.template_select_inc = 'select * from {src} where {rid} > {rid_value} order by {rid}'
        self.template_select_inc_null = 'select * from {src} order by {rid}'
        self.template_select_all = 'select * from {src}'
        self.template_select_rid = 'select max({rid}) from {src}'
        self.execute_batch_elems = connection_config.get('execute_batch_num_elems', 1000)
        self.active_cursor = None
        self.active_insert = None

    def insert_batch_(self, names, batch):
        with self.conn.cursor() as cur:
            if self.driver.paramstyle == 'pyformat':
                DBAPIEngine._insert_pyformat_optimized(cur, self.active_insert, names, batch,
                    batch_size=self.override_execute_batch_elems or self.execute_batch_elems)
            else:
                raise NotImplementedError('Support for drivers with paramstyle other than pyformat is not implemented!')
        

    def get_latest_rid(self, config):
        query = self.template_select_rid.format(
            src=DBAPIEngine._make_query_from_config(config),
            rid=config['rid']
        )
        with self.conn.cursor() as cur:
            cur.execute(query)
            res = cur.fetchall()
        if res is None or len(res) == 0:
            return None
        return res[0][0]

    def begin_incremental_fetch(self, config, min_rid):
        template = self.template_select_inc if min_rid else self.template_select_inc_null
        query = template.format(
            src=DBAPIEngine._make_query_from_config(config),
            rid=config['rid'],
            rid_value=min_rid
        )
        if self.active_cursor:
            self.active_cursor.close()

        self.active_cursor = self.conn.cursor()
        self.active_cursor.execute(query)

    def begin_full_fetch(self, config):
        query = self.template_select_all.format(
            src=DBAPIEngine._make_query_from_config(config)
        )
        if self.active_cursor:
            self.active_cursor.close()

        self.active_cursor = self.conn.cursor()
        self.active_cursor.execute(query)

    def begin_insert(self, config):
        name = config['table']
        if 'schema' in config:
            name = '{}.{}'.format(config['schema'], name)
        self.override_execute_batch_elems = config.get('execute_batch_num_elems')
        self.active_insert = name

    def fetch_batch(self, batch_size):
        if not self.active_cursor:
            raise Exception()
        keys = [x[0] if isinstance(x, tuple) else x.name for x in self.active_cursor.description]
        return keys, self.active_cursor.fetchmany(batch_size)

    def insert_batch(self, names, batch):
        self.insert_batch_(names, batch)
        self.conn.commit()

    def close(self):
        if self.active_cursor:
            self.active_cursor.close()
        self.conn.close()

add_engine_factory(DBAPIEngine.id, DBAPIEngine)



# Performance notes (based on simple test, without detailed hyperparameters and size comparisons):
# 1) pandas uses sqlalchemy with execute(table.insert(), data) -- set it as baseline
#    there are 3 option for insert method: None, 'multi' and custom (e.g. PG COPY)
# 2) None is versatile and fastest (100% time)
# 3) 'multi' is unexpectedly slower (700% time)
# 4) PG COPY is much faster (30% time)
# 5) None performance is attained using sqlalchemy.execute and passing list of dictionaries for bind params
# 6) Much faster alternative is insert values (), (), (), () using mogrify on python side ~50% time
#
# HENCE primary solution to test against pandas: mogrify inside python, insert with several simple executes

import functools

from .engine_base import BaseEngine
from .. import add_engine_factory

class SQLAlchemyEngine(BaseEngine):
    id = 'sqlalchemy'
    
    @staticmethod
    def _make_query_from_config( config):
        if 'query' in config:
            return '({}) t'.format(config['query'])
        if 'schema' in config:
            return '{}.{}'.format(config['schema'], config['table'])
        return '{}'.format(config['table'])

    def __init__(self, connection_config):
        import sqlalchemy #import only here when it will be actually used
        def make_table_(table_name, schema_name, col_names):
            meta = sqlalchemy.MetaData()
            table_kw = {}
            if schema_name:
                table_kw['schema'] = schema_name
            return sqlalchemy.Table(table_name, meta,
                        *[sqlalchemy.Column(x) for x in col_names],
                        **table_kw
                    )

        self.engine = sqlalchemy.create_engine(connection_config['conn-str'], **connection_config.get('sqlalchemy-config', {}))
        self.conn = self.engine.connect()
        self.template_select_inc = 'select * from {src} where {rid} > {rid_value} order by {rid}'
        self.template_select_inc_null = 'select * from {src} order by {rid}'
        self.template_select_all = 'select * from {src}'
        self.template_select_rid = 'select max({rid}) from {src}'
        self.make_query = sqlalchemy.text
        self.make_table = lambda table_name, schema_name, col_names: make_table_(table_name, schema_name, col_names)
        self.active_insert = None
        self.active_cursor = None

    def _execute(self, *args, **kwargs):
        try:
            return self.conn.execute(*args, **kwargs)
        except ConnectionError:
            self.conn = self.engine.connect()
            return self.conn.execute(*args, **kwargs)


    def get_latest_rid(self, config):
        query = self.make_query(self.template_select_rid.format(
            src=SQLAlchemyEngine._make_query_from_config(config),
            rid=config['rid']
        ))
        res = self._execute(query).fetchall()
        if res is None or len(res) == 0:
            return None
        return res[0][0]

    def begin_incremental_fetch(self, config, min_rid):
        template = self.template_select_inc if min_rid else self.template_select_inc_null
        query = self.make_query(template.format(
            src=SQLAlchemyEngine._make_query_from_config(config),
            rid=config['rid'],
            rid_value=min_rid
        ))
        self.active_cursor = self._execute(query)

    def begin_full_fetch(self, config):
        query = self.make_query(self.template_select_all.format(
            src=SQLAlchemyEngine._make_query_from_config(config)
        ))
        self.active_cursor = self._execute(query)

    def begin_insert(self, config):
        self.active_insert = functools.partial(self.make_table, table_name=config['table'], schema_name=config.get('schema'))

    def fetch_batch(self, batch_size):
        if not self.active_cursor:
            raise Exception()
        keys = list(self.active_cursor.keys())
        return keys, self.active_cursor.fetchmany(batch_size)

    def insert_batch(self, names, batch):
        self._execute(self.active_insert(col_names=names).insert(), [dict(zip(names, x)) for x in batch])

    def close(self):
        self.conn.close()
        self.engine.dispose()

add_engine_factory(SQLAlchemyEngine.id, SQLAlchemyEngine)


class BaseEngine:
    """
    Base abstract class for engine (connection) to data source/destination (e.g. database, message queue or pubsub).
    Purpose of this class is to provide getter/setter interface.
    """
    id = 'abstract'
    def __init__(self):
        pass

    def get_latest_rid(self, config):
        raise NotImplemented

    def begin_incremental_fetch(self, config, min_rid):
        raise NotImplemented

    def begin_full_fetch(self, config):
        raise NotImplemented

    def begin_insert(self, config):
        raise NotImplemented

    def fetch_batch(self, batch_size):
        raise NotImplemented

    def insert_batch(self, names, batch):
        raise NotImplemented

    def close(self):
        raise NotImplemented


"""
This module provides simple way to handle configs.
Key concept is having ability to override or generalize any given config.

So typical config structure should be as follows:
0. secret.key -- file containinig Fernet key
1. credentials.yaml.crypto -- encrypted credentials (or you may leave it unencrypted with credentials.yaml)
2. connections.yaml -- list of connections that can reference credentials {{level0.level1.level2}}
3. templates.yaml -- list of templates that are accessible for jobs
4. jobs.yaml -- list of jobs or pure job description
5. STDIN -- in form of level0.level1.level2=value, e.g. src.connection=ConName, src.table='asd'

So there must be a way to combine these multiple configs into single one. This could be done in two steps:
1. Unite all configs into single big config file: going from general to override/specific values
2. Use templating to update values of this final config
"""
import string
import functools
import copy
from typing import Any, Dict, List, Tuple, Union


def make_config(list_of_pairs : List[Tuple[str, Any]]) -> dict:
    """
    Convert list of pairs into structured config/dictionary.
    E.g. convert (a.b.c, 3) into {'a': {'b': {'c': 3}}}
    """
    if not list_of_pairs: #None, empty list and other should evaluate to empty dict
        return {}
    res = {}
    for k,v in list_of_pairs:
        keys = k.split('.')
        cur = res
        for x in keys[:-1]:
            if x not in cur:
                cur[x] = {}
            cur = cur[x]
        cur[keys[-1]] = v
    return res

def merge_handler_override(a, b):
    return b or a

def merge_handler_expand_or_override(a, b):
    if isinstance(a, list) and isinstance(b, list):
        return a + b
    return b or a

def merge_config(*args : dict, merge_handler=merge_handler_override) -> dict:
    """
    Merge several configs into single one. Later argument overrides the former.
    """
    def merge_dict(d1, d2, merge_handler):
        if isinstance(d1, dict) and isinstance(d2, dict):
            return {k: merge_dict(d1.get(k), d2.get(k), merge_handler) for k in set.union(set(d1), set(d2))}
        return merge_handler(d1, d2) #override with later        
    return functools.reduce(functools.partial(merge_dict, merge_handler=merge_handler), args, {})


def unflatten_config(config: Union[dict, list]) -> Union[dict, list]:
    if isinstance(config, list):
        return [unflatten_config(x) for x in config]
    elif isinstance(config, dict):
        cfg0 = {k:unflatten_config(v) for k,v in config.items() if '.' not in k}
        cfg1 = make_config([(k, unflatten_config(v)) for k,v in config.items() if '.' in k])
        return merge_config(cfg0, cfg1)
    else:
        return config


def flatten_config(config : dict, prefix : str = '') -> dict:
    """
    Convert config into flat map like a.b.c -> 1, a.b.d ->2, etc.
    """
    if not config: #None, empty dict and other should evaluate to empty dict
        return {}
    if not prefix:
        formatter = '{}'
    else:
        formatter = prefix + '.{}'
    res = {formatter.format(k): v for k,v in config.items()
                    if type(v) is not dict}
    res.update({k2: v2 for k, v in config.items()
                    if type(v) is dict
                    for k2, v2 in flatten_config(v, formatter.format(k)).items()})
    return res

def instantiate_templates(config: Union[dict, list], templates: dict, keywords = ['template', 'templates'], merge_handler = merge_handler_expand_or_override) -> Union[dict, list]:
    """
    Insert template values for `templates` dictionary into config, e.g.
    {
        "templates": ["a", "b"],
        "val": 3
    }
    {
        "a": {"va": 4, "vb": 5},
        "b": {"vb": 7, "vc": 8}
    }
    Should return
    {
        "templates": ["a", "b"],
        "va": 4,
        "vb": 7,
        "vc": 8,
        "val": 3
    }
    """
    if not config:
        return config
    if not templates:
        return config
    if isinstance(config, list):
        return [instantiate_templates(x, templates, keywords=keywords) for x in config]
    if not isinstance(config, dict):
        return config
    templates_to_use = []
    for k in keywords:
        if k in config:
            if isinstance(config[k], str):
                templates_to_use.append(config[k])
            elif isinstance(config[k], list) or isinstance(config[k], tuple):
                if not all(isinstance(x, str) for x in config[k]):
                    raise TypeError('Template value should be list of strings or single string, but got list with inconsistent types')
                templates_to_use += list(config[k])
            else:
                raise TypeError('Template value should be list of strings or single string, but got {}'.format(type(config[k])))
    if not templates_to_use: #exit early
        return config
    for t in templates_to_use:
        if t not in templates:
            raise ValueError('Template `{}` not present in templates dictionary'.format(t))
    
    result = merge_config(copy.deepcopy(config), *[templates[t] for t in templates_to_use], merge_handler=merge_handler)
    return {k: instantiate_templates(v, templates, keywords=keywords) for k,v in result.items()}
        
def expand_config(config: dict, field_from: str, field_to: str, merge_handler = merge_handler_expand_or_override) -> List[dict]:
    """
    Expand config into multiple items. E.g.
    {
        "values": [7, 10]
    }
    Will get expanded into
    [
        {
            "value": 7
        },
        {
            "value": 8
        }
    ]
    """
    tmp = config
    for p in field_from.split('.'):
        if p not in tmp:
            return [config]
        tmp = tmp[p]

    if not isinstance(tmp, list):
        raise TypeError('Config expansion can be made only on lists, but got {}'.format(type(tmp)))
    return [merge_config(copy.deepcopy(config), unflatten_config({field_to: v}), merge_handler=merge_handler) for v in tmp]

class TemplateDotted(string.Template):
    braceidpattern = r'[_a-z][_a-z0-9\.@\-]*'

def substitute_config(config : dict) -> dict:
    """
    Replace values like ${a.b.c} with value of a.b.c from this config
    """
    def replace_template(val, mapping):
        if isinstance(val, dict):
            return {k: replace_template(v, mapping) for k,v in val.items()}
        if isinstance(val, str):
            return TemplateDotted(val).substitute(mapping)
        return val
        
    if not config: #None, empty dict and other should evaluate to empty dict
        return {}
    flat_conf = flatten_config(config)
    while True:
        new_conf = replace_template(flat_conf, copy.deepcopy(flat_conf))
        if new_conf == flat_conf:
            break
        flat_conf = new_conf
    return replace_template(config, flat_conf)



__version__ = '0.0.11'

engine_factories = {}

def add_engine_factory(name, factory):
    global engine_factories
    engine_factories[name] = factory

def init_factory():
    from .engines.engine_sqlalchemy import SQLAlchemyEngine
    from .engines.engine_dbapi import DBAPIEngine
    from .engines.engine_pgcopy import PGCopyEngine
    add_engine_factory(SQLAlchemyEngine.id, SQLAlchemyEngine)
    add_engine_factory(DBAPIEngine.id, DBAPIEngine)
    add_engine_factory(PGCopyEngine.id, PGCopyEngine)

def create_engine(name, config):
    global engine_factories
    if name not in engine_factories:
        raise KeyError("Uknonwn engine: {}".format(name))
    return engine_factories[name](config)

init_factory()


import argparse
import logging
import sys
from typing import Dict, Optional, Union

from cryptography.fernet import Fernet
import yaml

from .config import make_config, merge_config, substitute_config
from .replication import full_copy, incremental_update
from . import create_engine, add_engine_factory, init_factory

logFormatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s]  %(message)s")
logger = logging.getLogger()
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setFormatter(logFormatter)
logger.addHandler(consoleHandler)
logger.setLevel(logging.DEBUG)

class StoreDictKeyPair(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(StoreDictKeyPair, self).__init__(option_strings, dest, nargs=nargs, **kwargs)
        
    def __call__(self, parser, namespace, values, option_string=None):
        my_dict = {}
        for kv in values:
            k,v = kv.split("=")
            my_dict[k] = v
        setattr(namespace, self.dest, my_dict)

def make_dbrep_argparser():
    parser = argparse.ArgumentParser(description='DataBase REPlication tool')
    subparsers = parser.add_subparsers(dest='cmd_main')
    parser_run = subparsers.add_parser('run', help='run replication between databases (and other entities)')
    parser_config = subparsers.add_parser('config', help='configure connections and templates (TBD), i.e. manage public data')
    parser_secret = subparsers.add_parser('secret', help='configure credentials and secrets, i.e. manage private data')

    parser_run.add_argument('-f', '--local', default='dbrep.yaml', help='Location of local configuration yaml (may reference global and credentials)')
    parser_run.add_argument('-g', '--globals', default=None, help='Location of global configuration yaml (may reference credentials)')
    parser_run.add_argument('-c', '--credential', default='dbrep.cred', help='Location of dbrep credentials file')
    parser_run.add_argument('-s', '--secret', default=None, help='Location of file with crypto-key')
    parser_run.add_argument('-r', '--run', default=None, help='Specify name of replication to run')
    parser_run.add_argument('-o', '--options', default=None, action=StoreDictKeyPair, nargs="*", metavar="KEY=VAL", help='Override options')

    parser_secret.add_argument('cmd', choices=['new', 'ls', 'rm', 'set'])
    parser_secret.add_argument('-s', '--secret', default=None, help='Location of file with crypto-key')
    parser_secret.add_argument('-c', '--credential', default='dbrep.cred', help='Location of dbrep credentials file')
    parser_secret.add_argument('-n', '--namespace', default=None, help='Location of file with crypto-key')
    parser_secret.add_argument('-o', '--options', default=None, action=StoreDictKeyPair, nargs="*", metavar="KEY=VAL", help='Configure values')

    return parser


def cli_dbrep():
    parser = make_dbrep_argparser()
    args = parser.parse_args()
    if args.cmd_main == 'run':
        secret = load_secret(args.secret)
        local_config = load_config(args.local, secret) or {}
        global_config = load_config(args.globals, secret)
        cred_config = load_config(args.credential, secret) or {}        
        options = make_config((args.options or {}).items())
        config = substitute_config(merge_config(global_config, cred_config, local_config, options))
        return run(config)
    elif args.cmd_main == 'secret':
        return manage_secrets(args)
    elif args.cmd_main == 'config':
        return manage_configs(args)
    else:
        raise NotImplementedError('Unexpected command in dbrep')

def load_secret(fname: Optional[str]) -> Optional[bytes]:
    if fname is None:
        return None
    with open(fname, 'r') as f: #propagate exception if invalid file
        return f.read().encode('utf-8')

def load_config(fname: Optional[str], secret: Optional[bytes] = None) -> Optional[Dict]:
    if fname is None:
        return None
    with open(fname, 'rb') as f:
        data = f.read()
    if secret is not None:
        frn = Fernet(secret)
        data = frn.decrypt(data)
    return yaml.safe_load(data.decode('utf-8'))

def manage_secrets(args):
    print('Invoke secrets with args: {}'.format(args))

def manage_configs(args):
    print('Invoke configs with args: {}'.format(args))


def make_engine(conn_config: Union[Dict, str], full_config: Dict):
    if isinstance(conn_config, str):
        if 'connections' not in full_config or conn_config not in full_config['connections']:
            raise KeyError('When specifying name of connection ({}), it should exist in `connections` section of config'.format(conn_config))
        config = full_config['connections'][conn_config]
    elif isinstance(conn_config, dict):
        config = conn_config
    else:
        raise TypeError('conn_config should be dict or str, but got {}'.format(conn_config))

    if not isinstance(config, dict):
        raise TypeError('Connection config should be dict with at least `engine` field, but got {}'.format(type(config)))
    if 'engine' not in config:
        raise ValueError('Connection config should contain at least `engine` field')

    return create_engine(config['engine'], config)

def run(config : Dict):
    init_factory()
    if 'run' not in config:
        if 'mode' not in config or 'src' not in config or 'dst' not in config:
            raise ValueError("Must specify `run` parameter: either name of replication from config, or dictionary specifying replication! Or specify mode, src and dst!")
        run_config = config
    elif isinstance(config['run'], str):
        run_config = config['replications'][config['run']] #Propagate exceptions
    elif isinstance(config['run'], dict):
        run_config = config['run']
    else:
        raise TypeError('Inapproapriate type of `run`: {}, while expected either dict or str'.format(type(config['run'])))
    
    if not isinstance(run_config, dict):
        raise TypeError('Run should be dict, but got {}!'.format(type(run_config)))
    if 'src' not in run_config or 'dst' not in run_config or 'mode' not in run_config:
        raise ValueError('Run should contain mode, src and dst')

    src_engine = make_engine(run_config['src']['conn'], config)
    dst_engine = make_engine(run_config['dst']['conn'], config)

    if run_config['mode'] == 'full-copy':
        return full_copy(src_engine, dst_engine, run_config)
    elif run_config['mode'] == 'incremental':
        return incremental_update(src_engine, dst_engine, run_config)
    else:
        raise ValueError("Unsupported mode: {}. Should be full-copy or incremental".format(run_config['mode']))




from typing import Any, Callable, List, Set, Tuple


def test_output_type(out) -> None:
    if out is None:
        raise TypeError('Output should be of type List[List[Any]], but got None')
    if not isinstance(out, list):
        raise TypeError('Output should be of type List[List[Any]], but got {}'.format(type(out)))
    if not all([isinstance(x, list) for x in out]):
        raise TypeError('Output should be of type List[List[Any]], but got {}'.format(set([type(x) for x in out if not isinstance(x, list)])))
    return   

def test_unique_key(out: List[List[Any]]) -> None:
    v0 = [x[0] for x in out] #Should raise if in some case length is 0
    if any([x is None for x in v0]):
        raise ValueError('Key in output should not be None!')
    if len(v0) != len(set(v0)):
        raise ValueError('Key should be unique, but got duplicates!')
    return

def merge_outputs(out1: List[List[Any]], out2: List[List[Any]]) -> List[List[Tuple[Any, Any]]]:
    def merge_row(r1, r2):
        if r1 is None and r2 is None:
            raise ValueError('Both rows can not be None!')
        if r1 is None or r2 is None:
            if r1 is None:
                return [(None, x) for x in r2]
            else:
                return [(x, None) for x in r1]
        if len(r1) != len(r2):
            raise ValueError('Both rows should have the same length!')
        return list(zip(r1, r2))

    all_keys = set.union(set(x[0] for x in out1), set(x[0] for x in out2))
    d1 = {x[0]: x for x in out1}
    d2 = {x[0]: x for x in out2}
    return [merge_row(d1.get(k), d2.get(k)) for k in all_keys]


def gather_exceptions(input: List[List[Tuple[Any, Any]]], test: Callable[[Tuple[Any, Any]], None]) -> List[List[Any]]:
    def gather_exc_(x: Tuple[Any, Any]) -> Any:
        try:
            test(x)
            return None
        except Exception as e:
            return type(e), e.args #if return just e it would not compare to itself and won't produce set
    return [[gather_exc_(x) for x in row] for row in input]

def agg_row_stats(input: List[List[Any]]) -> List[Set[Any]]:
    return [set(x for x in row if x is not None) for row in input]

def agg_col_stats(input: List[List[Any]]) -> List[Set[Any]]:
    if len(input) == 0:
        return []
    num_cols = len(input[0])
    return [set(r[i] for r in input if r[i] is not None) for i in range(num_cols)]

def agg_all_stats(input: List[List[Any]]) -> Set[Any]:
    return set(x for row in input for x in row if x is not None)

def run_tests(input: List[List[Tuple[Any, Any]]], test: Callable[[Tuple[Any, Any]], None], report_cols : bool = False):
    def format_exc_(exc):
        exc_type, exc_args = exc
        return '{}({})'.format(exc_type, '.'.join(exc_args))
    def format_exc_set_(excs):
        if len(excs) == 0:
            return '[]'
        return '[{}]'.format(', '.join(format_exc_(x) for x in excs))
    res = gather_exceptions(input, test)
    total = agg_all_stats(res)
    if len(total) == 0:
        return
    report = 'Triggered exceptions: {}'.format(format_exc_set_(total))
    if report_cols:
        cols = agg_col_stats(res)
        report += '\nExceptions by cols:\n{}'.format('\n'.join(format_exc_set_(x) for x in cols))
    raise Exception(report)

def test_elem_typing(input: Tuple[Any, Any]):
    if not isinstance(input, tuple):
        raise TypeError('Element should be tuple of length=2, but got type={}'.format(type(input)))
    if len(input) != 2:
        raise TypeError('Element should be tuple of length=2, but got length={}'.format(len(input)))
    return

def test_elem_none(input: Tuple[Any, Any]):
    a, b = input
    if a is None and b is None:
        return
    if a is None or b is None:
        raise ValueError('One elem is None, while other is not')
    return

def test_elem_type(input: Tuple[Any, Any]):
    a, b = input
    if type(a) != type(b):
        raise TypeError('Types differ: {} vs {}'.format(type(a), type(b)))
    return

def test_elem_value(input: Tuple[Any, Any]):
    a, b = input
    if a != b:
        raise ValueError('Values differ: {} vs {}'.format(a, b))
    return


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbrep', 'dbrep.engines']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.0', 'SQLAlchemy>=1.4.40,<2.0.0', 'cryptography>=37.0.4,<38.0.0']

entry_points = \
{'console_scripts': ['dbrep = dbrep.cli:cli_dbrep']}

setup_kwargs = {
    'name': 'dbrep',
    'version': '0.0.11',
    'description': 'Package to sync tables across DBs, i.e. EL in ELT/ETL',
    'long_description': "# dbrep\nPackage to sync tables across DBs, i.e. EL in ELT/ETL\n\n## What it is\nIt is tool for loading data from one DB and uploading into another. That is it. It is the simplest EL-tool.\n\n## What it is NOT\n- It is NOT orchestrator. You may use existing one, like Airflow, Prefect or Dagster.\n- It is NOT data transformation tool. You may use DBT or others.\n- It is NOT data ingestion tool -- it won't grab data from external resources.\n\n## Distinctive features\n### Simple\nIt does not try to solve miriad of problems, but nicely fits into existing data-engineering stack with (airflow, dbt and other tools).\n\n### Stateless\nIt does not store state of jobs or tables, besides connection configs.\n\n### Robust\nSince it is stateless there is only a handful of things that could go wrong.\n\n### Type-aware\nTypes are not lost in translation. I.e. if it was a *number(10,0)* in Oracle, it won't suddenly become a *Float64* in ClickHouse.",
    'author': 'Valentin  Stepanovich',
    'author_email': 'kolhizin@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)




import csv
from io import StringIO
from re import template

from .engine_dbapi import DBAPIEngine
from .. import add_engine_factory

class PGCopyEngine(DBAPIEngine):
    id = 'pgcopy'
    
    def __init__(self, connection_config):     
        connection_config['driver'] = 'psycopg2'
        super().__init__(connection_config)

    def insert_batch(self, names, batch):
        buffer = StringIO()

        cw = csv.writer(buffer, delimiter=';', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        cw.writerows(batch)
        buffer.seek(0)

        query = """
        copy {} from stdin csv delimiter ';'
        """.format(self.active_insert)

        with self.conn.cursor() as cur:
            cur.copy_expert(query, buffer)
        self.conn.commit()

add_engine_factory(PGCopyEngine.id, PGCopyEngine)